# ------------------------------------------------------------------------------
# Test Holidays
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from unittest.mock import Mock
from django.conf import settings
from django.test import TestCase, override_settings
from holidays import NZ, AU
from ls.joyous.models import CalendarPage
from ls.joyous.models import SimpleEventPage
from ls.joyous.holidays import Holidays
from ls.joyous.holidays.parser import parseHolidays, _parseSubdivisions
from .testutils import freeze_timetz, getPage

# ------------------------------------------------------------------------------
class TestHolidays(TestCase):
    @override_settings()
    def testNoSettings(self):
        del settings.JOYOUS_HOLIDAYS
        hols = Holidays()
        self.assertEqual(hols.simple, {})
        self.assertEqual(hols.srcs, [{}])
        self.assertEqual(hols.get(dt.date(1999,4,25)), "")

    def testNZSetting(self):
        hols = Holidays()
        self.assertEqual(hols.get(dt.date(1999,4,25)), "Anzac Day")

    @override_settings(JOYOUS_HOLIDAYS = None)
    def testSimple(self):
        hols = Holidays()
        hols.add(dt.date(1999,4,29), "HAPPY HAPPY")
        self.assertEqual(hols.get(dt.date(1999,4,29)), "HAPPY HAPPY")

    @override_settings(JOYOUS_HOLIDAYS = None)
    def testWorkalendar(self):
        class Woral:
            get_holiday_label = Mock(return_value="JOY JOY")
        woral = Woral()
        hols = Holidays()
        hols.register(woral)
        self.assertEqual(hols.srcs, [{}, woral])
        self.assertEqual(hols.get(dt.date(1999,4,30)), "JOY JOY")
        woral.get_holiday_label.assert_called_with(dt.date(1999,4,30))

    def testMultiHolidays(self):
        hols = Holidays()
        hols.add(dt.date(1999,1,1), "Gliffy")
        hols.add(dt.date(1999,1,1), "Whatnot")
        self.assertEqual(hols.get(dt.date(1999,1,1)),
                         "Gliffy, Whatnot, New Year's Day")

    @override_settings(JOYOUS_HOLIDAYS = None)
    def testNoNames(self):
        hols = Holidays()
        self.assertEqual(hols.names(), [])

    @freeze_timetz("2017-05-31")
    def testNZNames(self):
        hols = Holidays()
        self.assertEqual(hols.names(), [
                         "New Year's Day",
                         "Day after New Year's Day",
                         "New Year's Day (Observed)",
                         "Day after New Year's Day (Observed)",
                         'Wellington Anniversary Day',
                         'Auckland Anniversary Day',
                         'Nelson Anniversary Day',
                         'Waitangi Day',
                         'Waitangi Day (Observed)',
                         'Taranaki Anniversary Day',
                         'Otago Anniversary Day',
                         'Good Friday',
                         'Easter Monday',
                         'Southland Anniversary Day',
                         'Anzac Day',
                         'Anzac Day (Observed)',
                         "Queen's Birthday",
                         'South Canterbury Anniversary Day',
                         "Hawke's Bay Anniversary Day",
                         'Labour Day',
                         'Marlborough Anniversary Day',
                         'Canterbury Anniversary Day',
                         'Chatham Islands Anniversary Day',
                         'Westland Anniversary Day',
                         'Christmas Day',
                         'Boxing Day',
                         'Christmas Day (Observed)',
                         'Boxing Day (Observed)'])

    @override_settings(JOYOUS_HOLIDAYS = None)
    def testSimpleNames(self):
        hols = Holidays()
        hols.add(dt.date(2021,4,29), "HAPPY HAPPY")
        self.assertEqual(hols.names(), ["HAPPY HAPPY"])

    @override_settings(JOYOUS_HOLIDAYS = None)
    def testWorkalendarNames(self):
        class Woral:
            get_calendar_holidays = Mock(return_value=[(dt.date(1999,4,30),
                                                        "JOY JOY")])
        woral = Woral()
        hols = Holidays()
        hols.register(woral)
        self.assertEqual(hols.names(), ["JOY JOY"])

    def testAdd(self):
        ausHols = Holidays(None)
        ausHols.add(dt.date(2020,10,20), "Kangaroo Day")
        ausHols.register(AU())
        nzHols  = Holidays(None)
        nzHols.add(dt.date(2020,10,20), "Kiwi Day")
        nzHols.register(NZ())
        tasHols = ausHols + nzHols
        self.assertEqual(tasHols.get(dt.date(2020,10,20)),
                         "Kangaroo Day, Kiwi Day")
        self.assertEqual(len(tasHols.srcs), 3)
        self.assertIs(type(tasHols.srcs[0]), dict)
        self.assertIs(type(tasHols.srcs[1]), AU)
        self.assertIs(type(tasHols.srcs[2]), NZ)
        self.assertEqual(tasHols.names(), [
                         "New Year's Day",
                         "Day after New Year's Day",
                         "New Year's Day (Observed)",
                         "Day after New Year's Day (Observed)",
                         'Australia Day',
                         'Australia Day (Observed)',
                         'Waitangi Day',
                         'Waitangi Day (Observed)',
                         'Good Friday',
                         'Easter Monday',
                         'Anzac Day',
                         'Anzac Day (Observed)',
                         "Queen's Birthday",
                         'Kangaroo Day',
                         'Kiwi Day',
                         'Labour Day',
                         'Christmas Day',
                         'Boxing Day',
                         'Christmas Day (Observed)',
                         'Boxing Day (Observed)'])

# ------------------------------------------------------------------------------
class TestParser(TestCase):
    def testScotland(self):
        hols = parseHolidays("Scotland")
        self.assertEqual(hols.get(dt.date(2019,11,30)), "St. Andrew's Day")

    def testAllCountries(self):
        from ls.joyous.holidays.parser import _PYTHON_HOLIDAYS_MAP
        hols = parseHolidays("*")
        classes = [hol.__class__ for hol in hols.holidays if hol.country]
        self.assertCountEqual(classes, _PYTHON_HOLIDAYS_MAP.values())

    def testCountriesNE(self):
        hols = parseHolidays("*[NE]")
        self.assertEqual(hols.get(dt.date(2019,3,1)),
                         "Jahrestag der Ausrufung der Republik")
        self.assertEqual(hols.get(dt.date(2019,4,26)),
                         "Arbor Day")

    def testNorthIsland(self):
        hols = parseHolidays("NZ[NTL,AUK,HKB,TKI,WGN]")
        self.assertEqual(hols.get(dt.date(2020,1,20)),
                         "Wellington Anniversary Day")
        self.assertEqual(hols.get(dt.date(2020,1,27)),
                         "Auckland Anniversary Day")
        self.assertEqual(hols.get(dt.date(2020,3,9)),
                         "Taranaki Anniversary Day")
        self.assertEqual(hols.get(dt.date(2020,10,23)),
                         "Hawke's Bay Anniversary Day")

    def testNelson(self):
        hols = parseHolidays("NZ[Nelson]")
        self.assertEqual(hols.get(dt.date(2020,2,3)),
                         "Nelson Anniversary Day")

    def testInvalidCountry(self):
        self.assertIsNone(parseHolidays("Ruritania"))

    def testInvalidSubdivision(self):
        from holidays import UK
        self.assertEqual(_parseSubdivisions("ZZZ", UK), 0)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous Events
# ------------------------------------------------------------------------------

# This file is for backwards-compatibility (for migrations especially)

from .event_base import *
from .event_base import _get_default_timezone
from .one_off_events import *
from .recurring_events import *
from .events_api import *



# ------------------------------------------------------------------------------
# Joyous Holidays
# ------------------------------------------------------------------------------
import datetime as dt
from itertools import chain
from collections import defaultdict, OrderedDict
from django.conf import settings
from .parser import parseHolidays

class Holidays:
    """Defines what holidays are celebrated on what dates."""
    def __init__(self, holidaySetting="JOYOUS_HOLIDAYS"):
        self.setting = holidaySetting
        self.simple = {}
        self.srcs = [ self.simple ]
        self._parseSettings()

    def __add__(self, other):
        retval = Holidays(None)
        for date, value in chain(self.simple.items(), other.simple.items()):
            retval.add(date, value)
        for src in chain(self.srcs[1:], other.srcs[1:]):
            retval.register(src)
        return retval

    def _parseSettings(self):
        if self.setting:
            holidaySettings = getattr(settings, self.setting, "")
            if holidaySettings:
                hols = parseHolidays(holidaySettings)
                if hols is not None:
                    self.register(hols)

    def register(self, src):
        """Register a new source of holiday data."""
        self.srcs.append(src)

    def add(self, date, value):
        """Add a holiday to an individual date."""
        oldValue = self.simple.get(date)
        if oldValue:
            if oldValue not in value and value not in oldValue:
                self.simple[date] = "{}, {}".format(oldValue, value)
        else:
            self.simple[date] = value

    def get(self, date):
        """Get all the holidays that are celebrated on this date."""
        holidays = []
        for src in self.srcs:
            # get from python-holidays and other dict type srcs
            getHoliday = getattr(src, "get", None)
            if not getHoliday:
                # get from workalendar srcs
                getHoliday = getattr(src, "get_holiday_label")
            holiday = getHoliday(date)
            if holiday:
                holidays.extend(holiday.split(", "))
        holidays = list(OrderedDict.fromkeys(holidays))   # remove duplicates
        return ", ".join(holidays)

    def names(self):
        """Get a list of all the holiday names, sorted by month-day."""
        thisYear = dt.date.today().year
        popYears = list(range(thisYear - 1, thisYear + 10))
        holidays = defaultdict(list)
        for src in self.srcs:
            # populate python-holidays calendar
            populate = getattr(src, "_populate", None)
            if populate:
                for year in popYears:
                    populate(year)

            # get from python-holidays and other dict type srcs
            items = getattr(src, "items", None)
            if items:
                for date, names in items():
                    # holidays may have been concatenated together
                    for name in names.split(", "):
                        holidays[name].append(date)
            else:
                # get from workalendar srcs
                getHolidays = getattr(src, "get_calendar_holidays", None)
                if getHolidays:
                    for year in popYears:
                        for date, name in getHolidays(year):
                            holidays[name].append(date)

        # sort holidays by month-day with a preference for a more recent year
        mmddHolidays = []
        def moreRecent(date):
            delta2 = (date.year - thisYear) * 2
            if delta2 < 0:
                # slight preference for future years over past years
                delta2 = -delta2 + 1
            return delta2
        for name, dates in holidays.items():
            mostRecent = min(dates, key=moreRecent)
            mmddHolidays.append(((mostRecent.month, mostRecent.day), name))
        mmddHolidays.sort()
        retval = [name for mmdd, name in mmddHolidays]
        return retval

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Extended Cancellation Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from itertools import islice
from django.test import RequestFactory, TestCase
from django.contrib.auth.models import User
from django.utils import timezone
from wagtail.core.models import Page, PageViewRestriction
from wagtail.tests.utils.form_data import rich_text
from ls.joyous.models import CalendarPage
from ls.joyous.models import RecurringEventPage
from ls.joyous.models import ExtCancellationPage
from ls.joyous.utils.recurrence import Recurrence, WEEKLY, MONTHLY, MO, WE, FR
from .testutils import freeze_timetz, datetimetz


# ------------------------------------------------------------------------------
class Test(TestCase):
    @freeze_timetz("2020-03-18")
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3(r3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug      = "test-meeting",
                                        title     = "Test Meeting",
                                        repeat    = Recurrence(dtstart=dt.date(2010,1,1),
                                                               freq=WEEKLY,
                                                               byweekday=[MO,WE,FR]),
                                        time_from = dt.time(13),
                                        time_to   = dt.time(14,30))
        self.calendar.add_child(instance=self.event)
        self.shutdown = ExtCancellationPage(owner = self.user,
                                            overrides = self.event,
                                            cancelled_from_date=dt.date(2020,3,20),
                                            cancelled_to_date=dt.date(2020,6,1),
                                            cancellation_title="No Meeting during Shutdown")
        self.event.add_child(instance=self.shutdown)
        self.shutdown.save_revision().publish()

    @freeze_timetz("2021-01-04")
    def testInit(self):
        self.assertEqual(self.shutdown.title,
                         "Cancellation from Friday 20th of March to Monday 1st of June")
        self.assertEqual(self.shutdown.local_title,
                         "Cancellation from Friday 20th of March 2020 to Monday 1st of June 2020")
        self.assertEqual(self.shutdown.slug,  "2020-03-20-2020-06-01-cancellation")

    def testGetEventsByDay(self):
        shutdown0 = ExtCancellationPage(owner = self.user,
                                        overrides = self.event,
                                        cancelled_from_date=dt.date(2020,3,2),
                                        cancelled_to_date=dt.date(2020,3,6))
        self.event.add_child(instance=shutdown0)
        shutdown0.save_revision().publish()
        events = RecurringEventPage.events.byDay(dt.date(2020,3,1),
                                                 dt.date(2020,3,31))
        self.assertEqual(len(events), 31)
        evod = events[1]
        self.assertEqual(evod.date, dt.date(2020,3,2))
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 0)
        evod = events[8]
        self.assertEqual(evod.date, dt.date(2020,3,9))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        title, page, url = evod.days_events[0]
        self.assertEqual(title, "Test Meeting")
        evod = events[19]
        self.assertEqual(evod.date, dt.date(2020,3,20))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        title, page, url = evod.days_events[0]
        self.assertEqual(title, "No Meeting during Shutdown")
        self.assertIs(type(page), ExtCancellationPage)

    @freeze_timetz("2020-04-04")
    def testEventFutureExceptions(self):
        request = RequestFactory().get("/test")
        request.user = self.user
        request.session = {}
        exceptions = self.event._futureExceptions(request)
        self.assertEqual(len(exceptions), 1)
        shutdown = exceptions[0]
        self.assertEqual(shutdown.title,
                         "Cancellation from Friday 20th of March to Monday 1st of June")
        self.assertEqual(shutdown.cancellation_title, "No Meeting during Shutdown")
        self.assertEqual(shutdown._future_datetime_from, datetimetz(2020,4,6,13))

    @freeze_timetz("2020-08-31")
    def testPast(self):
        self.assertEqual(list(ExtCancellationPage.events.past()), [self.shutdown])
        self.assertEqual(ExtCancellationPage.events.past().count(), 1)
        self.assertEqual(ExtCancellationPage.events.future().count(), 0)

    @freeze_timetz("2020-05-04")
    def testFuture(self):
        self.assertEqual(list(ExtCancellationPage.events.future()), [self.shutdown])
        self.assertEqual(ExtCancellationPage.events.past().count(), 1)
        self.assertEqual(ExtCancellationPage.events.future().count(), 1)

    @freeze_timetz("2020-06-01 14:00:00")
    def testCurrent(self):
        self.assertEqual(list(ExtCancellationPage.events.current()), [self.shutdown])
        self.assertEqual(ExtCancellationPage.events.past().count(), 1)
        self.assertEqual(ExtCancellationPage.events.current().count(), 1)
        self.assertEqual(ExtCancellationPage.events.future().count(), 0)

    def testEventOccursOn(self):
        self.assertIs(self.event._occursOn(dt.date(2020, 5, 11)), False)

    @freeze_timetz("2020-05-31")
    def testEventNextDate(self):
        self.assertEqual(self.event.next_date, dt.date(2020,6,3))

    @freeze_timetz("2020-06-01")
    def testEventPastDatetime(self):
        self.assertEqual(self.event._past_datetime_from, datetimetz(2020,3,18,13,0))

    @freeze_timetz("2020-06-01")
    def testEventFutureDatetime(self):
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = self.event,
                                       cancelled_from_date = dt.date(2020,5,13))
        self.event.add_child(instance=shutdown)
        shutdown.save_revision().publish()
        self.assertIsNone(self.event._future_datetime_from)

    @freeze_timetz("2020-06-16")
    def testGetUpcomingEvents(self):
        event = RecurringEventPage(slug      = "lemon",
                                   title     = "Lemonade Stand",
                                   repeat    = Recurrence(dtstart=dt.date(2021,1,1),
                                                          freq=WEEKLY,
                                                          byweekday=[FR]),
                                   time_from = dt.time(11),
                                   time_to   = dt.time(13))
        self.calendar.add_child(instance=event)
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = event,
                                       cancelled_from_date = dt.date(2021,2,13),
                                       cancellation_title = "Gone fishing")
        event.add_child(instance=shutdown)
        shutdown.save_revision().publish()
        events = ExtCancellationPage.events.exclude(cancellation_title="")   \
                                      .upcoming().this()                     \
                                      .descendant_of(event)
        self.assertEqual(len(events), 1)
        title, page, url  = events[0]
        self.assertEqual(title, "Gone fishing")
        self.assertEqual(page._future_datetime_from, datetimetz(2021,2,19,11,0))
        self.assertEqual(url, "/events/lemon/2021-02-13--cancellation/")

    @freeze_timetz("2020-06-16")
    def testGetPastEvents(self):
        events = ExtCancellationPage.events.past().this()
        self.assertEqual(len(events), 1)
        title, page, url  = events[0]
        self.assertEqual(title, "No Meeting during Shutdown")
        self.assertEqual(page._past_datetime_from, datetimetz(2020,6,1,13,0))
        self.assertEqual(url, "/events/test-meeting/2020-03-20-2020-06-01-cancellation/")

    def testGetMyDates(self):
        dates = list(self.shutdown._getMyDates())
        self.assertEqual(dates,                                  [dt.date(2020,3,20),
                          dt.date(2020,3,23), dt.date(2020,3,25), dt.date(2020,3,27),
                          dt.date(2020,3,30), dt.date(2020,4, 1), dt.date(2020,4, 3),
                          dt.date(2020,4, 6), dt.date(2020,4, 8), dt.date(2020,4,10),
                          dt.date(2020,4,13), dt.date(2020,4,15), dt.date(2020,4,17),
                          dt.date(2020,4,20), dt.date(2020,4,22), dt.date(2020,4,24),
                          dt.date(2020,4,27), dt.date(2020,4,29), dt.date(2020,5, 1),
                          dt.date(2020,5, 4), dt.date(2020,5, 6), dt.date(2020,5, 8),
                          dt.date(2020,5,11), dt.date(2020,5,13), dt.date(2020,5,15),
                          dt.date(2020,5,18), dt.date(2020,5,20), dt.date(2020,5,22),
                          dt.date(2020,5,25), dt.date(2020,5,27), dt.date(2020,5,29),
                          dt.date(2020,6, 1)])

    def testGetMyRawDates(self):
        dates = list(self.shutdown._getMyRawDates(dt.date(2020,5,30),
                                                  dt.date(2020,6,10)))
        self.assertEqual(dates,
                         [dt.date(2020,5,30), dt.date(2020,5,31), dt.date(2020,6,1)])

    def testClosedOn(self):
        shutdown2 = ExtCancellationPage(owner = self.user,
                                        overrides = self.event,
                                        cancelled_from_date=dt.date(2020,9,2))
        self.event.add_child(instance=shutdown2)
        shutdown2.save_revision().publish()
        self.assertIs(shutdown2._closedOn(dt.date(2020, 9, 21)), True)
        self.assertIs(self.shutdown._closedOn(dt.date(2020, 3, 24)), True)
        self.assertIs(self.shutdown._closedOn(dt.date(2020, 6, 22)), False)

    def testStatus(self):
        self.assertEqual(self.shutdown.status, "cancelled")
        self.assertEqual(self.shutdown.status_text, "This event has been cancelled.")

    @freeze_timetz("2020-03-25 14:00")
    def testWhen(self):
        self.assertEqual(self.shutdown.when,
                         "Cancelled from Friday 20th of March to Monday 1st of June")

    @freeze_timetz("2020-03-25 14:00")
    def testWhenEver(self):
        event = RecurringEventPage(slug      = "OpQ",
                                   title     = "Orangepurple Quagga",
                                   repeat    = Recurrence(dtstart=dt.date(2020,1,1),
                                                          freq=MONTHLY,
                                                          byweekday=[FR(-1)]),
                                   time_from = dt.time(19))
        self.calendar.add_child(instance=event)
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = event,
                                       cancelled_from_date = dt.date(2020,4,1))
        event.add_child(instance=shutdown)
        shutdown.save_revision().publish()
        self.assertEqual(shutdown.when,
                         "Cancelled from Wednesday 1st of April until further notice")

    def testAt(self):
        self.assertEqual(self.shutdown.at.strip(), "1pm")

    @freeze_timetz("2020-03-25 14:00")
    def testCurrentDt(self):
        self.assertEqual(self.shutdown._current_datetime_from,
                         datetimetz(2020,3,25,13,0))

    @freeze_timetz("2020-03-25 14:00")
    def testFutureDt(self):
        self.assertEqual(self.shutdown._future_datetime_from,
                         datetimetz(2020,3,27,13,0))

    @freeze_timetz("2020-03-25 14:00")
    def testPastDt(self):
        self.assertEqual(self.shutdown._past_datetime_from,
                         datetimetz(2020,3,25,13,0))

    def testUnexplainedCancellation(self):
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = self.event,
                                       cancelled_from_date = dt.date(2019, 2, 8),
                                       cancellation_title   = "Restructure Pending",
                                       cancellation_details = "Keep it quiet")
        self.event.add_child(instance=shutdown)
        PASSWORD = PageViewRestriction.PASSWORD
        restriction = PageViewRestriction.objects.create(restriction_type = PASSWORD,
                                                         password = "s3cr3t",
                                                         page = shutdown)
        restriction.save()
        events = RecurringEventPage.events.byDay(dt.date(2019,2,1),
                                                 dt.date(2019,2,28))
        self.assertEqual(len(events), 28)
        evod = events[7]
        self.assertEqual(evod.date, dt.date(2019,2,8))
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 0)

    def testGroup(self):
        self.assertIsNone(self.shutdown.group)

    def testOverridesRepeat(self):
        self.assertEqual(self.shutdown.overrides_repeat, self.event.repeat)

    def testGetContext(self):
        request = RequestFactory().get("/test")
        context = self.shutdown.get_context(request)
        self.assertIn('overrides', context)

# ------------------------------------------------------------------------------
class TestPageForm(TestCase):
    Form = ExtCancellationPage.get_edit_handler().get_form_class()

    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@bar.test', 's3(r3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug      = "committee-meeting",
                                        title     = "Committee Meeting",
                                        repeat    = Recurrence(dtstart=dt.date(2017,1,1),
                                                               freq=MONTHLY,
                                                               byweekday=[MO(1), MO(3)]),
                                        time_from = dt.time(13),
                                        time_to   = dt.time(15,30))
        self.calendar.add_child(instance=self.event)
        self.shutdown = ExtCancellationPage(owner = self.user,
                                            overrides = self.event,
                                            cancelled_from_date=dt.date(2020,3,2),
                                            cancelled_to_date=dt.date(2020,3,6))
        self.event.add_child(instance=self.shutdown)
        self.shutdown.save_revision().publish()

    def testValid(self):
        page = ExtCancellationPage(owner = self.user)
        form = self.Form({'overrides':    self.event,
                          'cancelled_from_date':  "2020-03-10"},
                          instance = page, parent_page = self.event)
        self.assertTrue(form.is_valid())         # is_valid() calls full_clean()
        self.assertDictEqual(form.errors, {})

    def testExceptionAlreadyExists(self):
        page = ExtCancellationPage(owner=self.user)
        form = self.Form({'overrides':         self.event,
                          'cancelled_from_date':  "2020-03-02"},
                          instance=page,
                          parent_page=self.event)
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors,
                             {'cancelled_from_date':
                              ['There is already an extended cancellation for then']})

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Get Event Functions
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
import calendar
from django.test import RequestFactory, TestCase, override_settings
from django.contrib.auth.models import User, AnonymousUser, Group
from django.core.exceptions import (MultipleObjectsReturned, ObjectDoesNotExist,
                                    PermissionDenied)
from django.utils import timezone
from wagtail.core.models import Site, Page, PageViewRestriction
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import WEEKLY, MONTHLY, MO, TU, WE, FR, SU
from ls.joyous.models import GeneralCalendarPage
from ls.joyous.models import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, PostponementPage, CancellationPage, ExtraInfoPage)
from ls.joyous.models import (getAllEventsByDay, getAllEventsByWeek,
        getAllUpcomingEvents, getAllPastEvents, getGroupUpcomingEvents,
        getEventFromUid)
from ls.joyous.models import get_group_model
from .testutils import datetimetz

GroupPage = get_group_model()

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        self.home.add_child(instance=self.group)

        self.show = SimpleEventPage(owner = self.user,
                                    slug   = "pet-show",
                                    title  = "Pet Show",
                                    date      = dt.date(2013,1,5),
                                    time_from = dt.time(11),
                                    time_to   = dt.time(17,30),
                                    uid       = "29daefed-fed1-4e47-9408-43ec9b06a06d")
        self.calendar.add_child(instance=self.show)

        GROUPS = PageViewRestriction.GROUPS
        self.friends = Group.objects.create(name = "Friends")
        self.rendezvous = SimpleEventPage(owner = self.user,
                                          slug   = "rendezvous",
                                          title  = "Private Rendezvous",
                                          date      = dt.date(2013,1,10),
                                          uid       = "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.calendar.add_child(instance=self.rendezvous)
        self.rendezvous.save_revision().publish()
        restriction = PageViewRestriction.objects.create(restriction_type = GROUPS,
                                                         page = self.rendezvous)
        restriction.groups.set([self.friends])
        restriction.save()

        self.party = MultidayEventPage(owner = self.user,
                                       slug  = "allnighter",
                                       title = "All Night",
                                       date_from = dt.date(2012,12,31),
                                       date_to   = dt.date(2013,1,1),
                                       time_from = dt.time(23),
                                       time_to   = dt.time(3),
                                       uid       = "initiative+technology")
        self.calendar.add_child(instance=self.party)

        self.standup = RecurringEventPage(slug   = "test-meeting",
                                          title  = "Test Meeting",
                                          repeat = Recurrence(dtstart=dt.date(2013,1,1),
                                                              until=dt.date(2013,5,31),
                                                              freq=WEEKLY,
                                                              byweekday=[MO,WE,FR]),
                                          time_from = dt.time(13,30),
                                          time_to   = dt.time(16),
                                          uid       = "initiative+technology")
        self.group.add_child(instance=self.standup)

        self.postponement = PostponementPage(owner = self.user,
                                             slug  = "2013-01-09-postponement",
                                             title = "Postponement for Wednesday 16th of October",
                                             overrides = self.standup,
                                             except_date = dt.date(2013,1,16),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(2013,1,17),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.standup.add_child(instance=self.postponement)

        cancelTuesday = CancellationPage(owner = self.user,
                                         slug  = "2013-01-01-cancellation",
                                         title = "CancellationPage for Tuesday 1st of January",
                                         overrides = self.standup,
                                         except_date = dt.date(2013,1,1),
                                         cancellation_title   = "Meeting Cancelled")
        self.standup.add_child(instance=cancelTuesday)

    def testGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request,
                                   dt.date(2013,1,1), dt.date(2013,1,31),
                                   holidays=self.calendar.holidays)
        self.assertEqual(len(events), 31)
        evod1 = events[0]
        self.assertEqual(evod1.date, dt.date(2013,1,1))
        self.assertEqual(len(evod1.all_events), 1)
        self.assertEqual(len(evod1.days_events), 0)
        self.assertEqual(len(evod1.continuing_events), 1)
        evod10 = events[9]
        self.assertEqual(evod10.date, dt.date(2013,1,10))
        self.assertEqual(len(evod10.all_events), 0)

    def testAuthGetAllEventsByDay(self):
        self.request.user.groups.set([self.friends])
        events = getAllEventsByDay(self.request, dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod10 = events[9]
        self.assertEqual(evod10.date, dt.date(2013,1,10))
        self.assertEqual(len(evod10.all_events), 1)
        self.assertEqual(len(evod10.days_events), 1)
        event = evod10.days_events[0]
        self.assertEqual(event.title, "Private Rendezvous")

    def testGetAllEventsByWeek(self):
        weeks = getAllEventsByWeek(self.request, 2013, 1)
        self.assertEqual(len(weeks), 5)
        self.assertIsNone(weeks[0][0])
        self.assertIsNone(weeks[0][1])
        self.assertIsNone(weeks[4][5])
        self.assertIsNone(weeks[4][6])
        evod = weeks[2][4]
        self.assertEqual(evod.date, dt.date(2013,1,17))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)

    def testGetAllUpcomingEvents(self):
        today = timezone.localdate()
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from = today + dt.timedelta(days=1),
                                        date_to   = today + dt.timedelta(days=3),
                                        time_from = dt.time(17),
                                        time_to   = dt.time(10,30))
        self.calendar.add_child(instance=futureEvent)
        events = getAllUpcomingEvents(self.request, home=self.home,
                                      holidays=self.calendar.holidays)
        self.assertEqual(len(events), 1)
        title, event, url = events[0]
        self.assertEqual(title, "Tomorrow's Event")
        self.assertEqual(event.slug, "tomorrow")
        events0 = getAllUpcomingEvents(self.request)
        self.assertEqual(len(events0), 1)

    @override_settings(JOYOUS_UPCOMING_INCLUDES_STARTED = True)
    def testGetAllCurrentEvents(self):
        today = timezone.localdate()
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "yesterday",
                                        title = "Yesterday's Event",
                                        date_from = today - dt.timedelta(days=1),
                                        date_to   = today + dt.timedelta(days=3),
                                        time_from = dt.time(17),
                                        time_to   = dt.time(10,30))
        self.calendar.add_child(instance=futureEvent)
        events = getAllUpcomingEvents(self.request, home=self.home)
        self.assertEqual(len(events), 1)
        title, event, url = events[0]
        self.assertEqual(title, "Yesterday's Event")
        self.assertEqual(event.slug, "yesterday")
        events0 = getAllUpcomingEvents(self.request)
        self.assertEqual(len(events0), 1)

    def testGetAllPastEvents(self):
        events = getAllPastEvents(self.request)
        self.assertEqual(len(events), 5)
        self.assertEqual(events[0].title, "Test Meeting")
        self.assertEqual(events[1].title, "A Meeting")
        self.assertEqual(events[2].title, "Meeting Postponed")
        self.assertEqual(events[3].title, "Pet Show")
        self.assertEqual(events[4].title, "All Night")

    def testGetGroupUpcomingEvents(self):
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,1),
                                                            freq=WEEKLY,
                                                            byweekday=[TU]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = self.group)
        self.calendar.add_child(instance=meeting)
        memo = ExtraInfoPage(owner = self.user,
                             slug  = "plan-plan-extra-info",
                             title = "Extra Information Planning to Plan",
                             overrides = meeting,
                             except_date = meeting.next_date,
                             extra_title = "Gap Analysis",
                             extra_information = "Analyse your gaps")
        meeting.add_child(instance=memo)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].title, "Gap Analysis")
        self.assertEqual(events[1].title, "Planning to Plan")

    def testGetGroupUpcomingEventsDupGroup(self):
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,2),
                                                            freq=WEEKLY,
                                                            byweekday=[WE]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = self.group)
        self.group.add_child(instance=meeting)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[0].page.group, self.group)

    def testGetGroupUpcomingEvents2Groups(self):
        rival = GroupPage(slug = "initrode", title = "Initrode Corporation")
        self.home.add_child(instance=rival)
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,2),
                                                            freq=WEEKLY,
                                                            byweekday=[WE]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = rival)
        self.group.add_child(instance=meeting)
        events = getGroupUpcomingEvents(self.request, rival)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        # being a child of self.group trumps having group_page set
        self.assertEqual(events[0].page.group, self.group)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[0].page.group, self.group)

    def testGetEventFromUid(self):
        event = getEventFromUid(self.request, "29daefed-fed1-4e47-9408-43ec9b06a06d")
        self.assertEqual(event.title, "Pet Show")

    def testMultiGetEventFromUid(self):
        with self.assertRaises(MultipleObjectsReturned):
            getEventFromUid(self.request, "initiative+technology")

    def testMissingGetEventFromUid(self):
        with self.assertRaises(ObjectDoesNotExist):
            getEventFromUid(self.request, "d12971fb-e694-4a04-aba2-fb1a4a7166b9")

    def testAuthGetEventFromUid(self):
        with self.assertRaises(PermissionDenied):
            event = getEventFromUid(self.request, "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.request.user.groups.set([self.friends])
        event = getEventFromUid(self.request, "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.assertIsNotNone(event.title)
        self.assertEqual(event.title, "Private Rendezvous")

# ------------------------------------------------------------------------------
class TestTZ(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)

        self.night = RecurringEventPage(slug   = "pacnight",
                                        title  = "Pacific Night",
                                        repeat = Recurrence(dtstart=dt.date(2018,12,1),
                                                            count=1,
                                                            freq=MONTHLY,
                                                            byweekday=[SU(-1)]),
                                        time_from = dt.time(23,0),
                                        time_to   = dt.time(23,30),
                                        tz = pytz.timezone("Pacific/Pago_Pago"))
        self.calendar.add_child(instance=self.night)
        self.night.save_revision().publish()

    @timezone.override("Pacific/Kiritimati")
    def testExtremeTZGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request, dt.date(2019,1,1), dt.date(2019,1,31))
        self.assertEqual(len(events), 31)
        evod1 = events[0]
        self.assertEqual(evod1.date, dt.date(2019,1,1))
        self.assertEqual(len(evod1.all_events), 1)
        self.assertEqual(len(evod1.days_events), 1)
        self.assertEqual(evod1.days_events[0].title, "Pacific Night")

# ------------------------------------------------------------------------------
class TestNoCalendar(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.request.site = Site.objects.get(is_default_site=True)
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        self.home.add_child(instance=self.group)
        self.standup = RecurringEventPage(slug   = "test-meeting",
                                          title  = "Test Meeting",
                                          repeat = Recurrence(dtstart=dt.date(2013,1,1),
                                                              until=dt.date(2013,5,31),
                                                              freq=WEEKLY,
                                                              byweekday=[MO,WE,FR]),
                                          time_from = dt.time(13,30),
                                          time_to   = dt.time(16),
                                          uid       = "initiative+technology")
        self.group.add_child(instance=self.standup)
        self.postponement = PostponementPage(owner = self.user,
                                             slug  = "2013-01-09-postponement",
                                             title = "Postponement for Wednesday 16th of October",
                                             overrides = self.standup,
                                             except_date = dt.date(2013,1,16),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(2013,1,17),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.standup.add_child(instance=self.postponement)

    def testGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request, dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod2 = events[1]
        self.assertEqual(evod2.date, dt.date(2013,1,2))
        self.assertEqual(len(evod2.all_events), 1)
        self.assertEqual(len(evod2.continuing_events), 0)
        self.assertEqual(len(evod2.days_events), 1)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
